import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;


/**
 * WordCounter objects keep track of the number of times that each word (String) has been seen.
 * It implements the {@link Iterable} interface to allow clients to enumerate all encountered words. 
 */
public class WordCounter implements Iterable<String> {
    /**
     * Enable a quick lookup of the counter corresponding to a specific word.
     */
    private Map<String, Counter> wordMap; 
    
    /**
     * Constructs an empty <code>WordCounter</code>. 
     */
    public WordCounter() {
        wordMap = new HashMap<String, Counter>();
    }
        
    /**
     * Increments the counter corresponding to <code>word</code>.
     * @param word the string that is to be counted
     */
    public void count(String word) {
        // Find corresponding counter.
        Counter counter = wordMap.get(word);
        if (counter == null) {
            // This is a word that we haven't seen before.
            // Setup a new counter.
            counter = new Counter();
            wordMap.put(word, counter);
        }
        counter.increment();
    }

    /**
     * Returns how many times the word has been observed. It is ok to query
     * for a string that has never been observed.
     * @param word the query string
     * @return the number of times that <code>word</code> was observed.
     */
    public int getCount(String word) {
        Counter counter = wordMap.get(word);
        return counter == null ? 0 : counter.getCount();
    }
    
    /**
     * Allows iteration over the most-frequently seen words in order of decreasing frequency.
     * At most <code>howMany</code>
     * words are made available by the iterator. Returns an {@link Iterable} instead of directly an {@link Iterable}
     * so that it can be used in for-each loops. 
     * @param howMany a limit on the number of words that the iterator will visit.
     * @return an iterator that will yield at most <code>howMany</code> items
     */
    public Iterable<String> getMostFrequent(int howMany) {
        assert howMany >= 0;
        return new SortedView(howMany);
    }

    /**
     * Returns a default iterator. Allows iteration over all words in order of decreasing frequency.
     * @return an iterator that will yield as many items as there are in the map
     */
    public Iterator<String> iterator() {
        return new SortedView();
    }
    
    /**
     * Simple implementation of the {@link Iterator} and {@link Iterable} interface.  
     * It serves as a sorted view of the counter; words are returned in decreasing order of their counts.
     */
    private class SortedView implements Iterable<String>, Iterator<String>, Comparator<String> {

        /**
         * The words that were in WordCounter at the time that this SortedView was created,
         * ordered by decreasing frequency.
         */
        private String[] wordsInOrder;
        
        /**
         * The index of the next word to yield with <code>next()</code>. 
         */
        private int index;
        
        /**
         * The upper limit for <code>index</code>. When the limit is reached <code>hasNext()</code>
         * will return <code>false</code>.
         */
        private int maxIndex;
                       
        /**
         * Create a new view iterator that will yield at most <code>upperLimit<code> words. 
         * @param upperLimit the limit on the number of words that will be yielded 
         */
        SortedView(int upperLimit) {
            index = 0;
            maxIndex = upperLimit;
            wordsInOrder = new String[wordMap.size()];
            wordsInOrder = wordMap.keySet().toArray(wordsInOrder);
            Arrays.sort(wordsInOrder, this);
        }
        
        
        /**
         * Create a new view iterator with effectively no limit on the number of words. 
         */
        SortedView() {
            this(Integer.MAX_VALUE);
        }
        
        /* 
         * Dummy Iterable implementation: the iterator simply returns itself.
         * This allows the use of this iterator for-each loops.
         * @see java.lang.Iterable#iterator()
         */
        public Iterator<String> iterator() {
            return this;
        }

        /* 
         * Gets whether index is within permissible range. 
         * @see java.util.Iterator#hasNext()
         */
        public boolean hasNext() {
            return index < maxIndex && index < wordsInOrder.length;
        }

        public String next() {
            return wordsInOrder[index++];
        }

        public void remove() {
            throw new UnsupportedOperationException("Can't delete from sorted view.");
        }

        /** 
         * {@link Comparator} implementation: compare two strings by their respective counts.
         * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
         */
        public int compare(String word1, String word2) {
            return getCount(word2) - getCount(word1);
        }
        
    }
}
