import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Iterator;
import java.util.StringTokenizer;


/**
 * A convenience wrapper around a file or {@link Reader} that provides access
 * to the words in the file using the {@link Iterator} interface. This enables
 * files (and other streams) to be processes as a collection of strings.
 * 
 * This class could be replaced by using a {@link java.util.Scanner} instead.
 * In a production system, it would probably be better to use the standard Scanner.
 * It is implemented here by hand to illustrate how a {@link Reader} can be tokenized.
 */
public class WordIterator implements Iterable<String>, Iterator<String> {

    /**
     * The input that is being split into words.
     */
    private BufferedReader input;
    
    /**
     * The String that {@link #next()} will return next. 
     */
    private String next;
    
    /**
     * The line that is being processed.
     */
    private StringTokenizer tokenizer;
    
    /**
     * Hard-coded token list of characters that can separate words. 
     */
    private static final String NON_WORD_CHARS = " \n\r.:;,?!-\"\'()[]{}"; 
    
    /**
     * Create a new WordIterator that yields the words in a named text file 
     * @param filename the path of the text file that should be processed 
     * @throws FileNotFoundException - if the file cannot be opened
     */
    public WordIterator(String filename) throws FileNotFoundException {
        this(new FileReader(filename));
    }
    
    /**
     * Crate a new WordIterator that yields the words read from a {@link Reader}.
     * @param reader the stream of characters that should be split into words
     */
    public WordIterator(Reader reader) {
        input = new BufferedReader(reader);
        tokenizer = null;
        // obtain first word to initialize next
        getNextWord();
    }
    
    /*
     * Implement Iterator interface.
     * @see java.lang.Iterable#iterator()
     */
    public Iterator<String> iterator() {
        return this;
    }

    
    /* Implement Iterator interface.
     * @see java.util.Iterator#hasNext()
     */
    public boolean hasNext() {
        return next != null;
    }

    /* Implement Iterator interface.
     * @see java.util.Iterator#next()
     */
    public String next() {
        String prev = next;
        getNextWord();
        return prev;
    }

    /* Iterator requires this method, but it is optional.
     * @see java.util.Iterator#remove()
     */
    public void remove() {
        // Calling remove on this iterator doesn't make any sense.
        throw new UnsupportedOperationException("Cannot delete from word stream.");
    }

    /**
     * Fetches the next word from the tokenizer. If there aren't any left, then 
     * the next line is read. Updates next.
     */
    private void getNextWord() {
        if (tokenizer != null && tokenizer.hasMoreTokens()) {
            // just fetch the next word from tokenizer
            next = tokenizer.nextToken();
        } else
            // no more tokens in this line; need to fetch the next line.
            readNextLine();     
    }

    /**
     * Fetches the next line from the input. If there aren't any left or if an {@link IOException}
     * occurs, then next is set to null to indicate the end of the iteration. 
     */
    private void readNextLine() {
        String line = null;
        try {
            line = input.readLine();
        } catch (IOException exception) {
            // If we can't read any further, then just treat as end of file.
            // Thus, there is nothing to do.
        }
        if (line != null) {
            tokenizer = new StringTokenizer(line, NON_WORD_CHARS);
            // We have the next tokenizer. Now get a word.
            getNextWord();
        } else {
            // no more lines
            next = null;
            tokenizer = null;
        }
    }
}
